#include "PIKo_Module.h"

PIKo_Module::PIKo_Module(dWorldID world, dReal width, dReal height, dReal depth, dReal chassis_mass, dReal wheel_radius, dReal wheel_width, dReal wheel_mass)
{
	this->width = width;
	this->height = height;
	this->depth = depth;
	this->chassis_mass = chassis_mass;
	this->wheel_radius = wheel_radius;
	this->wheel_width = wheel_width;
	this->wheel_mass = wheel_mass;

	chassis = new Chassis(world, width, height, depth, chassis_mass);

	r1Wheel = new Wheel(world, wheel_radius, wheel_width, wheel_mass);
	l1Wheel = new Wheel(world, wheel_radius, wheel_width, wheel_mass);
	r2Wheel = new Wheel(world, wheel_radius, wheel_width, wheel_mass);
	l2Wheel = new Wheel(world, wheel_radius, wheel_width, wheel_mass);

	r1Joint = dJointCreateHinge2(world, 0);
	l1Joint = dJointCreateHinge2(world, 0);
	r2Joint = dJointCreateHinge2(world, 0);
	l2Joint = dJointCreateHinge2(world, 0);
}


void PIKo_Module::setPosition(dReal x, dReal y, dReal z)		//Allways use setRotation before setPosition
{
	// constants (0.005) is distance between wheel and body, not implemented as variable 
	dVector3 p0 = {x, y, z, 1.0f}, temp;
	dVector3 p1Left = {0.0f, depth/2+0.005, -height/2, 0.0f};			//body-fixed coordinates of left wheel
	dVector3 p1Right = {0.0f, -depth/2-0.005, -height/2, 0.0f};		//body-fixed coordinates of right wheel
	dVector3 p2Left = {0.0f, depth/2+0.005, height/2, 0.0f};			//body-fixed coordinates of left upper wheel 
	dVector3 p2Right = {0.0f, -depth/2-0.005, height/2, 0.0f};		//body-fixed coordinates of right lower wheel

	chassis->setPosition(x, y, z);
	
	//set wheel positions
	dMultiply0(temp, chassis->getRotation(), p1Left, 3, 3, 1);		//rotate body-fixed vector
	p1Left[0] = temp[0] + p0[0];							
	p1Left[1] = temp[1] + p0[1];										//Now left wheel coordinates in inertial frame
	p1Left[2] = temp[2] + p0[2];
	l1Wheel->setPosition(p1Left[0], p1Left[1], p1Left[2]);

	dMultiply0(temp, chassis->getRotation(), p1Right, 3, 3, 1);		//rotate body-fixed vector
	p1Right[0] = temp[0] + p0[0];
	p1Right[1] = temp[1] + p0[1];										//Now right wheel coordinates in inertial frame
	p1Right[2] = temp[2] + p0[2];
	r1Wheel->setPosition(p1Right[0], p1Right[1], p1Right[2]);

	dMultiply0(temp, chassis->getRotation(), p2Left, 3, 3, 1);		//rotate body-fixed vector
	p2Left[0] = temp[0] + p0[0];							
	p2Left[1] = temp[1] + p0[1];										//Now left wheel coordinates in inertial frame
	p2Left[2] = temp[2] + p0[2];
	l2Wheel->setPosition(p2Left[0], p2Left[1], p2Left[2]);

	dMultiply0(temp, chassis->getRotation(), p2Right, 3, 3, 1);		//rotate body-fixed vector
	p2Right[0] = temp[0] + p0[0];
	p2Right[1] = temp[1] + p0[1];										//Now right wheel coordinates in inertial frame
	p2Right[2] = temp[2] + p0[2];
	r2Wheel->setPosition(p2Right[0], p2Right[1], p2Right[2]);

	//Attach joints
	dVector3 axis10 = {0, 0, 1, 1.0f};
	dVector3 axis20 = {0, 1, 0, 1.0f};
	dVector3 axis1, axis2;
	dMultiply0(axis1, chassis->getRotation(), axis10, 3,3,1);
	dMultiply0(axis2, chassis->getRotation(), axis20, 3,3,1);

	dJointAttach(r1Joint, chassis->getBody(), r1Wheel->getBody());
	const dReal *a = dBodyGetPosition (r1Wheel->getBody());
	dJointSetHinge2Anchor (r1Joint,a[0],a[1],a[2]);
	dJointSetHinge2Axis1 (r1Joint,axis1[0],axis1[1],axis1[2]);
	dJointSetHinge2Axis2 (r1Joint,axis2[0],axis2[1],axis2[2]);
	dJointSetHinge2Param (r1Joint,dParamHiStop,0); //Originally 0
	dJointSetHinge2Param (r1Joint,dParamLoStop,0);
	dJointSetHinge2Param (r1Joint,dParamHiStop,0);
	dJointSetHinge2Param (r1Joint,dParamCFM,0);
	dJointSetHinge2Param (r1Joint,dParamERP,0.8);

	dJointAttach(l1Joint, chassis->getBody(), l1Wheel->getBody());
	const dReal *b = dBodyGetPosition (l1Wheel->getBody());
	dJointSetHinge2Anchor (l1Joint,b[0],b[1],b[2]);
	dJointSetHinge2Axis1 (l1Joint,axis1[0],axis1[1],axis1[2]);
	dJointSetHinge2Axis2 (l1Joint,axis2[0],axis2[1],axis2[2]);
	dJointSetHinge2Param (l1Joint,dParamHiStop,0);
	dJointSetHinge2Param (l1Joint,dParamLoStop,0);
	//dJointSetHinge2Param (l1Joint,dParamHiStop,1.3);
	dJointSetHinge2Param (l1Joint,dParamCFM,0);
	dJointSetHinge2Param (l1Joint,dParamERP,0.8);

	dJointAttach(r2Joint, chassis->getBody(), r2Wheel->getBody());
	a = dBodyGetPosition (r2Wheel->getBody());
	dJointSetHinge2Anchor (r2Joint,a[0],a[1],a[2]);
	dJointSetHinge2Axis1 (r2Joint,axis1[0],axis1[1],axis1[2]);
	dJointSetHinge2Axis2 (r2Joint,axis2[0],axis2[1],axis2[2]);
	dJointSetHinge2Param (r2Joint,dParamHiStop,0);
	dJointSetHinge2Param (r2Joint,dParamLoStop,0);
	//dJointSetHinge2Param (r2Joint,dParamHiStop,1.3);
	dJointSetHinge2Param (r2Joint,dParamCFM,0);
	dJointSetHinge2Param (r2Joint,dParamERP,0.8);

	dJointAttach(l2Joint, chassis->getBody(), l2Wheel->getBody());
	b = dBodyGetPosition (l2Wheel->getBody());
	dJointSetHinge2Anchor (l2Joint,b[0],b[1],b[2]);
	dJointSetHinge2Axis1 (l2Joint,axis1[0],axis1[1],axis1[2]);
	dJointSetHinge2Axis2 (l2Joint,axis2[0],axis2[1],axis2[2]);
	dJointSetHinge2Param (l2Joint,dParamHiStop,0);
	dJointSetHinge2Param (l2Joint,dParamLoStop,0);
	//dJointSetHinge2Param (l2Joint,dParamHiStop,1.3);
	dJointSetHinge2Param (l2Joint,dParamCFM,0);
	dJointSetHinge2Param (l2Joint,dParamERP,0.8); //TODO: Check if CFM/ERP tweaks help wheel damage
}

void PIKo_Module::setRotation(const dMatrix3 R)			//Allways use setRotation before setPosition
{
	chassis->setRotation(R);
	
	dMatrix3 rot1, rot2, rot10, rot20;
	dRFromAxisAndAngle(rot1, 1, 0, 0, M_PI/2);				//Additional rotation to make the wheels upright
	dRFromAxisAndAngle(rot2, 1, 0, 0, -M_PI/2);				//Additional rotation to make the wheels upright

	dMultiply0(rot10, R, rot1, 3,3,3);
	r1Wheel->setRotation(rot10);
	r2Wheel->setRotation(rot10);

	dMultiply0(rot20, R, rot1, 3,3,3);
	l1Wheel->setRotation(rot20);
	l2Wheel->setRotation(rot20);
}

void PIKo_Module::addToSpace(dSpaceID space)
{
	chassis->addToSpace(space);
	r1Wheel->addToSpace(space);
	l1Wheel->addToSpace(space);
	r2Wheel->addToSpace(space);
	l2Wheel->addToSpace(space);
}

dBodyID PIKo_Module::getBody()
{
	return chassis->getBody();
}

const dReal * PIKo_Module::getPosition()
{
	return chassis->getPosition();
}

/* Gets the euler angles from ODE matrix */
/*void get_euler(const dReal *matrix, dReal &kx, dReal &ky, dReal &kz)
{
	// Euler angles: http://web.archive.org/web/20041029003853/http:/www.j3d.org/matrix_faq/matrfaq_latest.html#Q37
	ky = asin(matrix[8]); // tr_matrix[2]
	C = cos (ky);
	if (fabs(C) < 0.005) {
		tr_x = matrix[10] / C;
		tr_y = -matrix[9] / C; // tr_matrix[6]
		kx = atan2(tr_y, tr_x);
		tr_x = matrix[0] / C;
		tr_y = -matrix[4] / C; // tr_matrix[1]
		kz = atan2(tr_y, tr_x);
	} else {
		kx = 0;
		tr_x = matrix[5];
		tr_y = matrix[1]; // tr_matrix[4]
		kz = atan2(tr_y, tr_x);
	}
	kx = (kx * 180.0)/M_PI;
	ky = (ky * 180.0)/M_PI; // Angle up-down
        kz = (kz * 180.0)/M_PI; // Angle left-right
}*/

const dReal * PIKo_Module::getRotation() {
       dReal *v;
       v = new dReal [3];
       //get_euler(dBodyGetRotation(chassis->getBody()),v[0],v[1],v[2]);
       return v;
}

dReal * PIKo_Module::getDimensions()
{
	return chassis->getDimensions();
}

void PIKo_Module::addToViewer(Viewer *v)
{
	if (v==NULL)
		return;
	chassis->addToViewer(v);
	r1Wheel->addToViewer(v);
	l1Wheel->addToViewer(v);
	r2Wheel->addToViewer(v);
	l2Wheel->addToViewer(v);
}
