#include "Simulator.h"

#define GRAVITY 9.81

#ifdef _MSC_VER
#pragma warning(disable:4244 4305)  // for VC++, no precision loss complaints
#endif

dWorldID world;
dJointGroupID contactgroup;

//function called by dSpaceCollide and dSpaceCollide when two objects are near, and potentially colliding
static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
	//if one of the objects is a space itself then recall function recursively
	if (dGeomIsSpace (o1) || dGeomIsSpace (o2)) {
		dSpaceCollide2 (o1,o2,data,&nearCallback);
		return;
	}

	const int N = 30; //max number of contactpoints between two surfaces
	dContact contact[N];
	int n = dCollide (o1,o2,N,&contact[0].geom,sizeof(dContact));
	if (n > 0) {
		for (int i=0; i<n; i++) {
			contact[i].surface.mode = dContactSlip1 | dContactSlip2 |
				dContactSoftERP | dContactSoftCFM | dContactApprox1;
			contact[i].surface.mu = dInfinity;
			contact[i].surface.slip1 = 0.01;
			contact[i].surface.slip2 = 0.01;
			contact[i].surface.soft_erp = 0.2;
			contact[i].surface.soft_cfm = 0.00001;
			dJointID c = dJointCreateContact (world,contactgroup,&contact[i]); //Create and join a contact joint
			dJointAttach (c, dGeomGetBody(contact[i].geom.g1),
					dGeomGetBody(contact[i].geom.g2));
		}
	}
}

Simulator::Simulator(dReal pos_x, dReal pos_y, dReal pos_z, double timestep, bool visual) {
	this->timestep = timestep;
	this->visual = visual;
	dInitODE2(0);
	world = dWorldCreate();
	space = dHashSpaceCreate(0);

	//Subspace for all actors, currently only train
	actor_space = dHashSpaceCreate(space);
	dSpaceSetCleanup(actor_space, 0);

	//Subspace for static objects such as ground and vents
	static_space = dHashSpaceCreate(space);
	dSpaceSetCleanup(static_space, 0);

	contactgroup = dJointGroupCreate(0);
	dWorldSetGravity(world, 0, 0, -GRAVITY);
	//Create ground
	ground = dCreatePlane(static_space, 0, 0, 1, 0);

	//Create all actors
	train = new Train(world, 6);
	train->setPosition(pos_x, pos_y, pos_z);
	train->addToSpace(actor_space);
	train->resetDistance();

	if (visual)
		viewer = new Viewer(800, 600);
	else
		viewer = NULL;

	//Create all static objects
	//	dMatrix3 R;

	vents = new EATestVentStruct();
	vents->setPosition(-1,0,0.15);
	vents->addToSpace(static_space);

	//Add all objects to viewer
	if (visual) {
		train->addToViewer(viewer);
		vents->addToViewer(viewer);
	}
	bool log = false; // Shoud be set from constructor input
	if (log) {
		logfile.open ("log.txt");
	}

	running = true;
}

void Simulator::step(void)
{

	//Collide objects from the actor_space to objects form the static_space
	dSpaceCollide2 ((dGeomID)actor_space, (dGeomID)static_space, 0, &nearCallback);

	//dSpaceCollide (space,0,&nearCallback); //Collide objects in space

	dWorldStep (world, timestep);

	//Empty the contactgroup before next time_step
	dJointGroupEmpty(contactgroup);
	if (viewer!=NULL)
		running = viewer->draw();

	if (logfile.is_open()) {
		logfile << getTrain()->getDistanceNose((dGeomID)getStaticSpace()) << "\t";
		logfile << getTrain()->getTravelledDistance() << std::endl;
	}
}

Simulator::~Simulator(void)
{
	//Terminate program
	dJointGroupDestroy (contactgroup);
	dSpaceDestroy (space);
	dWorldDestroy (world);
	dCloseODE();

	if (logfile.is_open())
		logfile.close();
	delete viewer;
}
