#include "Train.h"
#include <iostream>

Train::Train(dWorldID world)
{
	Train(world, CARTS_DEFAULT);	
}

Train::Train(dWorldID world, int carts)
{
	// Set number of carts in train.
	num_carts = carts;

	// Allocate memory:
	for (int i=0;i<num_carts;i++) {
		module.push_back(new PIKo_Module(world, CHASS_W, CHASS_H, CHASS_D, CHASS_M, WHEEL_R, WHEEL_W, WHEEL_M));
	}

	// Create vertical and horizontal joints and brackets
	dJointGroupID trainJoints = dJointGroupCreate(0);

	for (int i = 0; i < num_carts-1; i++) {
		vertjoint.push_back(dJointCreateHinge(world,trainJoints));
		horjoint.push_back(dJointCreateHinge(world,trainJoints));
		bracket.push_back(new Bracket(world, BRACK_W, BRACK_H, BRACK_D, BRACK_M));
		jointAngleSetPointVert.push_back(0);
		jointAngleSetPointHor.push_back(0);
	}

	// Initialize measurement ray
	dReal measureMaxLength = 100;
	noseRay = dCreateRay (NULL, measureMaxLength);
	dGeomRaySetClosestHit(noseRay, true);

	// Initialize PID
	lastError = 0;
	integral = 0;
}

void Train::setPosition(dReal x, dReal y, dReal z)
{
	module[0]->setPosition(x, y, z);
	dReal *bodyDimension = module[0]->getDimensions();
	const dReal *currentBodyPos;
	const dReal *modulePos;
	const dReal *bracketPos;
	dMatrix3 R;
	dRFromAxisAndAngle(R, 0, 0, 1, 0);

	// Place carts
	for (int i = 0; i < getNumCarts() ; i++) {	
		module[i]->setRotation(R);
		module[i]->setPosition(x + i*(bodyDimension[0] + CART_SPACE), y, z);
	}

	for (int i = 0; i < getNumJoints() ; i++) {	
		// Place brackets
		currentBodyPos = module[i]->getPosition();
		bracket[i]->setRotation(R);
		bracket[i]->setPosition(currentBodyPos[0] + bodyDimension[0]/2 + CART_SPACE*0.5, currentBodyPos[1], currentBodyPos[2]);

		// Attach joints, one less joint than number of carts
		dJointAttach(vertjoint[i],module[i]->getBody(),bracket[i]->getBody());
		dJointAttach(horjoint[i], bracket[i]->getBody(), module[i+1]->getBody());

		// Create joint feedback structures.
		dJointSetFeedback(vertjoint[i], new dJointFeedback);
		dJointSetFeedback(horjoint[i], new dJointFeedback);

		modulePos = dBodyGetPosition (module[i]->getBody());
		dJointSetHingeAnchor (vertjoint[i],modulePos[0] + bodyDimension[0]/2 + CART_SPACE*0.25,modulePos[1],modulePos[2]);
		dJointSetHingeAxis (vertjoint[i],0,0,1); //axis relative body 1

		// Degrees constraints on joint
		dJointSetHingeParam (vertjoint[i], dParamLoStop, -ANGLE_CONSTR);
		dJointSetHingeParam (vertjoint[i], dParamHiStop, ANGLE_CONSTR);

		bracketPos = dBodyGetPosition (bracket[i]->getBody());
		// Anchor position must be set in world coordinates
		dJointSetHingeAnchor (horjoint[i],modulePos[0] + bodyDimension[0]/2 + CART_SPACE*0.75,modulePos[1],modulePos[2]);
		dJointSetHingeAxis (horjoint[i],0,1,0); //axis relative body 1

		// Degrees constraints on joint
		dJointSetHingeParam (horjoint[i], dParamLoStop, -ANGLE_CONSTR);
		dJointSetHingeParam (horjoint[i], dParamHiStop, ANGLE_CONSTR);
	}

	// Store initial position for later reset
	initial_x = x;
	initial_y = y;
	initial_z = z;
}

void Train::addToSpace(dSpaceID space)
{
	for (int i = 0; i < num_carts; i++) module[i]->addToSpace(space);
	for (int i = 0; i < num_carts-1; i++) bracket[i]->addToSpace(space);
}

void Train::addToViewer(Viewer *v) {
	if (v==NULL)
		return;
	for(int i = 0; i < num_carts; i++)
		module[i]->addToViewer(v);

	for(int i = 0; i < num_carts-1; i++)
		bracket[i]->addToViewer(v);
}

// Reset train to initial conditions
// TODO: Angles on bracets isn't reset properly
void Train::resetTrain(void)
{
	setPosition(initial_x, initial_y, initial_z);

	setMotorSpeedAll(0);
	setMotorTorqueAll(0);
	setJointTorqueHorizontal(0);
	setJointTorqueVertical(0);
	setJointAngleHorizontal(0);
	setJointAngleVertical(0);

	integral = 0;
	lastError = 0;

	resetDistance();
}

// Find forces on pressure sensors in joints
void Train::getJointForces(int jointNumber, double *up, double *down, double *left, double *right)
{
	if (jointNumber >= getNumJoints())
		return;

	dJointFeedback *fb_hor = dJointGetFeedback(horjoint[jointNumber]);
	dJointFeedback *fb_vert = dJointGetFeedback(vertjoint[jointNumber]);

	dVector3 result;

	// Contribution from force along x-axis, split between all sensors
	dBodyVectorFromWorld(bracket[jointNumber]->getBody(), fb_vert->f2[0], fb_vert->f2[1], fb_vert->f2[2], result);
	*up = result[0]/4;
	*down = result[0]/4;
	*left = result[0]/4;
	*right = result[0]/4;

	// Contribution from force along x-axis, split between all sensors
	dBodyVectorFromWorld(bracket[jointNumber]->getBody(), fb_hor->f1[0], fb_hor->f1[1], fb_hor->f1[2], result);
	*up += result[0]/4;
	*down += result[0]/4;
	*left += result[0]/4;
	*right += result[0]/4;

	// Contribution from torque
	dBodyVectorFromWorld(bracket[jointNumber]->getBody(), fb_vert->t2[0], fb_vert->t2[1], fb_vert->t2[2], result);
	*up += result[0]/FORK_LENGTH;
	*down += -result[0]/FORK_LENGTH;
	*left += result[2]/FORK_LENGTH;
	*right += -result[2]/FORK_LENGTH;

	dBodyVectorFromWorld(bracket[jointNumber]->getBody(), fb_hor->t1[0], fb_hor->t1[1], fb_hor->t1[2], result);
	*up += result[0]/FORK_LENGTH;
	*down += -result[0]/FORK_LENGTH;
	*left += result[2]/FORK_LENGTH;
	*right += -result[2]/FORK_LENGTH;
}

void Train::setMotorSpeed(int moduleNumber, dReal speed)
{
	if (moduleNumber >= num_carts) return;

	// Calculate angular velocity from linear
	speed /= WHEEL_R;

	dJointSetHinge2Param (module[moduleNumber]->r1Joint,dParamVel2,speed);		//Set reference speed of hinge joints
	dJointSetHinge2Param (module[moduleNumber]->r1Joint,dParamFMax2,MAX_MOTOR_TORQUE);		//Set maximum force excertable by the joint motors

	dJointSetHinge2Param (module[moduleNumber]->l1Joint,dParamVel2,speed);
	dJointSetHinge2Param (module[moduleNumber]->l1Joint,dParamFMax2,MAX_MOTOR_TORQUE);

	dJointSetHinge2Param (module[moduleNumber]->r2Joint,dParamVel2,-speed);
	dJointSetHinge2Param (module[moduleNumber]->r2Joint,dParamFMax2,MAX_MOTOR_TORQUE);

	dJointSetHinge2Param (module[moduleNumber]->l2Joint,dParamVel2,-speed);
	dJointSetHinge2Param (module[moduleNumber]->l2Joint,dParamFMax2,MAX_MOTOR_TORQUE);
}

void Train::setMotorSpeedAll(dReal speed)
{
	for(int i = 0; i < num_carts; i++) {
		setMotorSpeed(i, speed);
	}
}

// Set torque on wheels. Should not be used, use setMotorSpeed instead
void Train::setMotorTorque(int moduleNumber, dReal torque)
{
	//Check if this module exists:
	if (moduleNumber >= num_carts) return;	

	if (abs(torque) > abs(MAX_MOTOR_TORQUE)) {
		if (torque > MAX_MOTOR_TORQUE)
			torque = MAX_MOTOR_TORQUE;
		else
			torque = -MAX_MOTOR_TORQUE;
	}

	// Apply torque to wheel axis, zero around suspension axis
	dJointAddHinge2Torques(module[moduleNumber]->r1Joint, 0.0f, torque);
	dJointAddHinge2Torques(module[moduleNumber]->r2Joint, 0.0f, torque);
	dJointAddHinge2Torques(module[moduleNumber]->l1Joint, 0.0f, torque);
	dJointAddHinge2Torques(module[moduleNumber]->l2Joint, 0.0f, torque);
}

void Train::setMotorTorqueAll(dReal torque)
{
	for(int i = 0; i < num_carts; i++) {
		setMotorTorque(i, torque);
	}
}

// Set joints to gived angle using PID regulator
void Train::setJointAngle(int jointNumber, int direction, dReal angle)
{
	//Check if this module exists:
	if (jointNumber >= getNumJoints()) return;

	dJointID *joint;

	dReal currentAngle, currentAngleRate;
	if (direction == 0) {
		joint = &vertjoint[jointNumber];
		jointAngleSetPointVert[jointNumber] = angle;
	} else {
		joint = &horjoint[jointNumber];
		jointAngleSetPointHor[jointNumber] = angle;
	}

	currentAngle = dJointGetHingeAngle(*joint);
	currentAngleRate = dJointGetHingeAngleRate(*joint);

	//PID-regulator
		dReal k_p = 3.0, k_d = 0.0, k_i = 0.0;
		dReal error = angle - currentAngle;
		dReal dt = 0.05; //Timestep HardCoded
		integral += error*dt;
		dReal derivative = (error - lastError)/dt;
		dReal result = (k_p*error) + (k_i*integral) + (k_d*derivative);
		lastError = error;
	
	//setJointTorque(jointNumber, direction, result);
	setJointVelocity(jointNumber, direction, result);
}

void Train::setJointAngleHorizontal(dReal angle)
{
	for(int i = 0; i < num_carts-1; i++) {
		setJointAngle(i, 0, angle);
	}
}

void Train::setJointAngleVertical(dReal angle)
{
	for(int i = 0; i < num_carts-1; i++) {
		setJointAngle(i, 1, angle);
	}
}

// Set torques on joints, for internal use in setJointAngle
void Train::setJointTorque(int jointNumber, int direction, dReal torque)
{
	//Check if this module exists:
	if (jointNumber >= getNumJoints()) return;	

	if (torque > MAX_JOINT_TORQUE)
		torque = MAX_JOINT_TORQUE;
	else if (torque < -MAX_JOINT_TORQUE)
		torque = -MAX_JOINT_TORQUE;

	if (direction == 0)
		dJointAddHingeTorque(vertjoint[jointNumber], torque);
	else
		dJointAddHingeTorque(horjoint[jointNumber], torque);
}

void Train::setJointTorqueHorizontal(dReal torque)
{
	for(int i = 0; i < num_carts-1; i++) {
		dJointAddHingeTorque(horjoint[i], torque);
	}
}

void Train::setJointTorqueVertical(dReal torque)
{
	for(int i = 0; i < num_carts-1; i++) {
		dJointAddHingeTorque(vertjoint[i], torque);
	}
}

void Train::setJointVelocity(int jointNumber, int direction, dReal velocity)
{
	//Check if this module exists:
	if (jointNumber >= getNumJoints())
		return;

	if (velocity > MAX_JOINT_VELOCITY)
		velocity = MAX_JOINT_VELOCITY;
	else if (velocity < -MAX_JOINT_VELOCITY)
		velocity = -MAX_JOINT_VELOCITY;

	if (direction == 0) {
		dJointSetHingeParam(vertjoint[jointNumber], dParamVel, velocity);
		dJointSetHingeParam(vertjoint[jointNumber], dParamFMax, MAX_JOINT_TORQUE);
	} else {
		dJointSetHingeParam(horjoint[jointNumber], dParamVel, velocity);
		dJointSetHingeParam(horjoint[jointNumber], dParamFMax, MAX_JOINT_TORQUE);
	}

}

// Find distance from first module to closest static object with direction
// pitch and yaw.
double Train::getDistance(dGeomID space, dReal pitch, dReal yaw)
{
	const dReal *position, *rotation;
	position = module[0]->chassis->getPosition();
	rotation = module[0]->chassis->getRotation();
	dVector3 forward = {-1, 0, 0}, forward_rot, forward_yaw;
	dMatrix3 rot_pitch, rot_yaw;
	dRFromAxisAndAngle(rot_pitch, 0, 1, 0, pitch);
	dMultiply0(forward_rot, rot_pitch, forward, 3, 3, 1);
	dRFromAxisAndAngle(rot_yaw, 0, 0, 1, yaw);
	dMultiply0(forward_yaw, rot_yaw, forward_rot, 3, 3, 1);

	dBodyVectorToWorld(module[0]->chassis->getBody(), forward_yaw[0], forward_yaw[1], forward_yaw[2], forward);
	dGeomRaySet(noseRay, position[0], position[1], position[2], forward[0], forward[1], forward[2]);
	dContactGeom contact[1];
	dCollide(noseRay, space, 1, contact, sizeof(dContactGeom));

	// ODE sets depth of rays that doesn't hit anyting to DBL_MIN, set
	// them to DBL_MAX instead.
	if (contact[0].depth > 0.001)
		return (double)contact[0].depth;
	else
		return (double)DBL_MAX;
}

double Train::getDistanceNose(dGeomID space)
{
	return getDistance(space, 0, 0);
}

// Ultrasound-like distance measurement, returns shortest distance within the
// opening angle.
double Train::getSonar(dGeomID space, double openingAngle)
{
	int numPoints = 10;
	double minDistance = DBL_MAX, distance;
	for (double i = -openingAngle/2; i < openingAngle/2; i+=openingAngle/numPoints) {
		for (double j = -openingAngle/2; j < openingAngle/2; j+=openingAngle/numPoints) {
			distance = getDistance(space, i, j);
			if (distance < minDistance)
				minDistance = distance;
		}
	}
	return minDistance;
}

// Gives an array of distances in a horizontal plane with given number of
// points and openingAngle.
void Train::getLaserScanner1D(dGeomID space, int points, double openingAngle, double* result)
{
	for (int i = 0; i < points; i++) {
		result[i] = getDistance(space, i*openingAngle/points - openingAngle/2, 0);
	}
}

// 2D version of getLaserScanner1D
void Train::getTimeOfFlightCam(dGeomID space, int verticalPoints, int horizontalPoints, double verticalOpeningAngle, double horizontalOpeningAngle, double **result)
{
	for (int i = 0; i < horizontalPoints; i++) {
		for (int j = 0; j < verticalPoints; j++) {
			result[i][j] = getDistance(space, i*horizontalOpeningAngle/horizontalPoints - horizontalOpeningAngle/2, j*verticalOpeningAngle/verticalPoints - verticalOpeningAngle/2);
		}
	}
}

// Extrace IMU-like data from ODE
// TODO: Estimate linear and angular acceleration from ODE data.
void Train::getIMU(double **position, double **direction, double **linearVelocity, double
**angularVelocity, double **linearAcc, double **angularAcc)
{
       (*position) = (double *)module[0]->chassis->getPosition();
       dReal *v; v = new dReal [3];
       dBodyVectorToWorld(module[0]->chassis->getBody(), 1, 0, 0, v);
       (*direction) = (double *)v;
       (*linearVelocity) = (double *)dBodyGetLinearVel(module[0]->chassis->getBody());
       (*angularVelocity) = (double *)dBodyGetAngularVel(module[0]->chassis->getBody());
}

double Train::getJointAngle(int jointNumber, int direction)
{
	//Check if this module exists:
  	if (jointNumber >= getNumJoints()) return 0;

	if (direction == 0)
		return (double)dJointGetHingeAngle(vertjoint[jointNumber]);
	else
		return (double)dJointGetHingeAngle(horjoint[jointNumber]);
}

double Train::getJointAngleSetPoint(int jointNumber, int direction)
{
	if (jointNumber >= getNumJoints())
		return 0;

	if (direction == 0)
		return (double)jointAngleSetPointVert[jointNumber];
	else
		return (double)jointAngleSetPointHor[jointNumber];
}

// Finds distance travelled along the path of the first cart
double Train::getTravelledDistance(void)
{
	const dReal *position;
	position = module[0]->chassis->getPosition();
	distance += sqrt(pow(position[0]-lastPosition[0], 2) + pow(position[1]-lastPosition[1], 2) + pow(position[2]-lastPosition[2], 2));
	for (int i = 0; i < 3; i++) {
		lastPosition[i] = position[i];
	}
	return distance;
}

void Train::resetDistance(void)
{
	const dReal *position;
	position = module[0]->chassis->getPosition();
	for (int i = 0; i < 3; i++) {
		lastPosition[i] = position[i];
	}
	distance = 0;
	wheelDistance = 0;
}

// Find distance travelled from wheel rotation.
// More realistic measurement than getTravelledDistance, but is affected by
// wheelspin.
double Train::getWheelDistance(double timestep)
{
	const dReal *velocity;
	velocity = dBodyGetAngularVel(module[0]->r1Wheel->getBody());
	double total_velocity = sqrt(velocity[0]*velocity[0] + velocity[1]*velocity[1] + velocity[2]*velocity[2]);
	wheelDistance += WHEEL_R*total_velocity*timestep;
	return wheelDistance;
}

double Train::getHeigth(int moduleNumber)
{
	if (moduleNumber >= num_carts) return 0;

	const dReal *position;
	position = module[moduleNumber]->chassis->getPosition();
	return (double)position[2];
}
